/*
  endian.c : functions specific to the endianness, to read and write in
  motorola format.
*/

#include "endian.h"

/* if in doubt, do it the slow way */
#ifndef BIGENDIAN
#define BIGENDIAN 0
#endif

#if BIGENDIAN
long get_long(void *addr)
{
  long *laddr = (long *) addr;
  return (*laddr);
}

void set_long(void *addr, long value)
{
  long *laddr = (long *) addr;
  *laddr = value;
}

short get_word(void *addr)
{
  short *saddr = (short *) addr;
  return (*saddr);
}

void set_word(void *addr, short value)
{
  short *saddr = (short *) addr;
  *saddr = value;
}

#else 

long get_long(void *addr)
{
  unsigned char *uaddr = (unsigned char *) addr;
  return (uaddr[0] << 24) + (uaddr[1] << 16) + (uaddr[2] << 8) + uaddr[3];
}

void set_long(void *addr, long value)
{
  unsigned char *uaddr = (unsigned char *) addr;
  uaddr[0] = value >> 24;
  uaddr[1] = value >> 16;
  uaddr[2] = value >> 8;
  uaddr[3] = value;
}

short get_word(void *addr)
{
  unsigned char *uaddr = (unsigned char *) addr;
  return (uaddr[0] << 8) + uaddr[1];
}

void set_word(void *addr, short value)
{
  unsigned char *uaddr = (unsigned char *) addr;
  uaddr[0] = value >> 8;
  uaddr[1] = value;
}

#endif

