/*
 *   MX -> ST : MINIX 68k 1.5 format to Atari ST gemdos executable.
 *
 *  usage : mxst [<source> [<destination>]]
 *           ( source and dest default to a.out and a.prg respectively. )
 */

/*
  what it does :
  - read and convert header.
  - add jump to start position to text. 
    modify start of reloc info accordingly.
  - copy text + data, relocating if jump code added;
  - copy symb table with conversion.
  - copy reloc verbatim
  
*/

#include <stdio.h>
#include <stdlib.h>

#include "stheader.h"
#include "mxheader.h"
#include "endian.h"

static int verbose = 0;

long dosymbols(FILE *from, FILE *to, long ssize);

/* returns the actual length */

long dosymbols(FILE *from, FILE *to, long ssize)
{
  struct nlist b;
  struct tos_symbol a;
  int i;
  size_t count, len;
  long value;
  long ssize2 = 0;
  
  long nsyms = ssize / sizeof(b);
  while(nsyms > 0) {
    nsyms--;
    count = sizeof(b);
    len = fread(&b, 1, count, from);
    if(len != count) goto fail;
    
    /* todo, examine symbol types */
    value = GET_LONG(b.n_value);
    for(i = 0 ; i < 8 ; i++) {
      a.sym_name[i] = b.n_name[i];
    }
    SET_WORD(a.sym_type, SYM_PLAIN); 
    SET_LONG(a.sym_value, value);
    
    count = sizeof(a);
    len = fwrite(&a, 1, count, to);
    if(len != count) goto fail;
    ssize2 += count;
  }
  return ssize2;
fail:
  perror("");
  exit(1);
}


/* return codes :
   2 = bad header
   3 = 
   4 = could not access to files
   5 = bad reloc
*/
int mxtost(char *src_name, char *dest_name)
{
  FILE *from, *to;
	
  struct tos_header t ;
  struct mx_header m ;
  size_t count, len ;
  unsigned char buf[1024] ;
  char jump_code[8];
  unsigned char reloc_buf[1024];
  int err;
    
  long tsize, dsize, bsize, ssize, ssize2, entry ;
  long essize2; /* estimated tos symb size */
  
  int jump_code_size = 0;

/* read mx header */

  from = fopen(src_name, "rb");
  if(from == NULL) return 4;

  len = fread((char *)&m, 1, sizeof(m), from);
  if(len != sizeof(m)) return 4;

  if(m.a_cpu != 0x04) return 2;
  if(m.a_flags != 0x10) return 2;
  if(GET_WORD(&m.a_magic) != 0x0301) return 2;

  /* skip extra header parameters */
  if(m.a_hdrlen > sizeof(m)) {
    count = m.a_hdrlen - sizeof(m);
    len = fread(buf, 1, count, from);
    if(len != count) return 4;
  }


  tsize = GET_LONG(m.a_text);
  dsize = GET_LONG(m.a_data);
  bsize = GET_LONG(m.a_bss);
  ssize = GET_LONG(m.a_syms);
  entry = GET_LONG(m.a_no_entry);

  /* now, emit st header */

  SET_WORD(t.prg_magic, PRG_MAGIC);
  if(entry < 0 || entry & 1) {
    fprintf(stderr, "entry negative or odd\n");
    return 2;
  }
  
  /* determine jump code */
  if(entry == 0) {
    jump_code_size = 0;
  } else if(entry <= 0x7E) {
    jump_code_size = 2;
    /* bra.s _start */
    SET_WORD(jump_code, 0x6000 + (short) entry);
  } else if(entry <= 0x7FFE) {
    jump_code_size = 4;
    /* bra _start */
    SET_WORD(jump_code, 0x6000);
    SET_WORD(jump_code+2, (short) entry);
  } else {
    jump_code_size = 6;
 /*     0:  4EF9 <start>    jmp _start
  */
    SET_WORD(jump_code, 0x4EF9);
    SET_LONG(jump_code+2, entry + 6);
  }
  
  SET_LONG(t.prg_tsize, tsize + jump_code_size);
  SET_LONG(t.prg_dsize, dsize);
  SET_LONG(t.prg_bsize, bsize);
  /* put estimated symb size now. */
  essize2 = (ssize / sizeof(struct nlist)) * sizeof(struct tos_symbol);
  SET_LONG(t.prg_ssize, essize2);
  SET_LONG(t.prg_res1, 0L);
  SET_LONG(t.prg_flags, 0L);
  SET_WORD(t.prg_abs, 0);
  
  to = fopen(dest_name, "wb");
  if(to == NULL) return 4;
   
  count = sizeof(t);
  len = fwrite(&t, 1, count, to);
  if(count != len) return 4;

  /* if needed, copy jump */
  if(jump_code_size) {
    count = jump_code_size;
    len = fwrite(&t, 1, count, to);
    if(count != len) return 4;
  }
    

  /* now, output text and data segts */

  if(jump_code_size == 0) {
    /* simple copy */
    len = tsize + dsize ;
    while(len > 0) {
      count = 1024;
      if(count > len)
        count = len;
      count = fread(buf, 1, count, from);
      count = fwrite(buf, 1, count, to);
      len -= count;
    }
  } else {
    /* must relocate text+data */
    long window_start;
    int  window_len;
    long adr ;
    unsigned char *a;
    unsigned char *p;	/* pointer to next reloc byte */
    int n;      /* number of relocs remaining in buffer */
    int c;
    long reloc_offset; 
    FILE *reloc; 
    reloc_offset = tsize + dsize + ssize + m.a_hdrlen;
    reloc = fopen(src_name, "rb");
    if(reloc == NULL) return 4;
    err = fseek(reloc, reloc_offset, SEEK_SET);
    if(err) return 4;
    
    
    count = fread(reloc_buf, 1, 1024, reloc);
    if(count < 4) return 4;
    
    window_start = 0;
    
    /* load first window */
    window_len = 1024;
    if(window_len > tsize + dsize) {
      window_len = tsize + dsize;
    }
    len = window_len;
    count = fread(buf, 1, len, from);
    if(count != len) return 1;
    window_len -= 4;

    adr = GET_LONG(reloc_buf);
    if(adr == 0) {
      goto no_more_reloc;
    }

    n = count - 4;
    p = reloc_buf + 4;
    
    /* following code taken from mm/exec.c */
    for (;;) {			/* once per relocation */
      /* here, relocate long at address adr 
       * (adr = 0 at beginning of text)
       * (was:) *((long *)adr) += off;
       */
       
      while(adr >= window_start + window_len) {
      	/* ship out current window */
      	len = window_len;
      	count = fwrite(buf, 1, len, to);
        if(count != len)  return 4;
        window_start += window_len;
      	
      	/* copy additional end byte to the beginning of the buffer */
      	buf[0] = buf[window_len];
      	buf[1] = buf[window_len+1];
      	buf[2] = buf[window_len+2];
      	buf[3] = buf[window_len+3];
      	
      	/* load next window */
      	window_len = 1024 - 4;
      	if(window_start + window_len > tsize + dsize) {
      	  window_len = tsize + dsize - window_start;
      	}
      	if(window_len <= 0) {
      	  goto no_more_reloc;
      	}
      	len = window_len;
      	count = fread(buf+4, 1, len, from);
      	if(count != len) return 4;
      }
      a = buf + (adr - window_start);
      if(verbose) {
      	fprintf(stderr, "relocating offset %lx\n", adr);
      }
      SET_LONG(a, GET_LONG(a) + jump_code_size);
      for (;;) {		/* once per byte */
        if (--n < 0) {
          p = reloc_buf;
          count = 1024;
          count = fread(reloc_buf, 1, count, reloc);
          if (count <= 0)
            return 4;	/* error */
          n = count - 1;
        }
        c = *p++ & 0xFF;
        if (c != 1)
          break;
        adr += 254;
      }          
      if (c == 0)
        break;
      if (c & 1)
        return 5;	/* error */
      adr += c;
    }
 
no_more_reloc:
    window_len += 4;
    /* ship out the window, and copy the rest */
    while(window_len > 0) {
      len = window_len;
      count = fwrite(buf, 1, len, to);
      if(count != len)  return 4;
      window_start += window_len;
      window_len = 1024;
      if(window_start + window_len > tsize + dsize) {
        window_len = tsize + dsize - window_start;
      }
      len = window_len;
      count = fread(buf, 1, len, from);
      if(count != len) return 4;
    }
    
    fclose(reloc);
  }

  /* convert symb table if any */

  if(ssize) {
    ssize2 = dosymbols(from, to, ssize);
  }

  /* finally, output reloc info */

  if(jump_code_size == 6) {
    long adr ;
    unsigned char *p;	

    /* need to insert offset 2 in front of the relocs */
    len = 4;
    len = fread(buf, 1, len, from);
    if(len != 4) return 4;
    adr = GET_LONG(buf);
    SET_LONG(buf, 2L);
    if(adr == 0) { 
      /* no more reloc */
      buf[4] = 0;
      len = 5;
      len = fwrite(buf, 1, len, to);
      if(len != 5) return 4;
      goto end_output_reloc;
    } else {
      p = buf+4;
      len = 4;
      while(adr > 254) {
        *p++ = 1;
        len++;
        if(len >= 1024) {
          len = 1024;
          len = fwrite(buf, 1, len, to);
          if(len != 1024) return 4;
          p = buf;
          len = 0;
        }
        adr -= 254;
      }
      if(adr > 0) {
        *p = adr;
        len++;
      }
      if(len) {
        count = fwrite(buf, 1, len, to);
        if(count != len) return 4;
      }
    }
  }
      
      
  do {
    len = 1024;
    len = fread(buf, 1, len, from);
    count = fwrite(buf, 1, len, to);
    if(count != len) return 4;
  } 
  while(len > 0);

end_output_reloc:
  
  if(essize2 != ssize2) {
    int err;
    /* patch back the header */
    err = fseek( to, 0L, SEEK_SET);
    if(err) {
      fprintf(stderr, "cannot patch back the header.\n");
      return 1;
    }
    SET_LONG(t.prg_ssize, ssize2);
    fwrite((char *)&m, 1, sizeof(m), to);
  }    

  fclose(to);
  fclose(from);

  return 0;
}


int main(int argc, char **argv)
{
  char *src_name ;
  char *dest_name ;
  int err;

  if(argc == 1) {
    src_name = "a.out" ;
    dest_name = "a.prg" ;
  } else if(argc == 2) {
    src_name = argv[1] ;
    dest_name = "b.out" ;
  } else if(argc == 3) {
    src_name = argv[1] ;
    dest_name = argv[2] ;
  } else {
    goto usage;
  }

  err = mxtost(src_name, dest_name);
  switch(err) {
  case 0:
    exit(0);
  case 4:
    perror(argv[0]);
    exit(1);
  default:
    fprintf(stderr, "error number %d\n", err);
    exit(1);
  }
  
usage:
  fprintf(stderr, "usage: %s [<src> [<dest>]]\n", argv[0]);
  fprintf(stderr, "<src> and <dest> default to a.out and a.prg respectively\n");
  exit(1);
}
