/*
 * st-header.h : GEMDOS header for executable programs
 *
 * updated with info on the symbol table format.
 */

/*

  shorts and longs have been replaced by unsigned char[]. 
  use byte-order independent macros to access fields.
 */


struct tos_header {
  unsigned char prg_magic[2] ;
#define PRG_MAGIC 0x601a
  unsigned char prg_tsize[4] ;
  unsigned char prg_dsize[4] ;
  unsigned char prg_bsize[4] ;
  unsigned char prg_ssize[4] ;   /* sym table length, 0 for no sym table */
  unsigned char prg_res1[4] ;    /* reserved, 0 */
  unsigned char prg_flags[4] ;   /* 0 for private memory */
#define PRG_FLAGS 7
  unsigned char prg_abs[2] ;     /* 0 if reloc info */
#define PRG_ABS 0
} ;

struct tos_symbol {
  unsigned char sym_name[8] ;
  unsigned char sym_type[2] ;
  /* type value for ordinary symbols */
#define SYM_PLAIN ((short)0xA200)    
  /* type value for symbol whose name overlaps over the next symbol slot */
#define SYM_CONTD ((short)0xA248)    
  unsigned char sym_value[4] ;
} ;


/* order of sections in the executable file :

  - header
  - .text
  - .data
  - symbol table
  - reloc

  relocation info in GEMDOS format :
  - only longs can be relocated ;
  - first long is offset from beginning of file of first long to relocate,
    or zero if no relocation. 
  - the rest of the file are bytes with following meaning :
  -- 0 : end of relocation.
  -- 1 : advance 254 bytes.
  -- 2 <= count <= 254 (count even) : advance count bytes then fixup the
     long there.
  -- other : error.
 
  Always put reloc info (prg_abs == 0), and if no reloc just put a long zero
  after .data section, or some loaders (some tos versions ?) will bug.

  From the rare few examples I've seen, and litterature I've read, the symbol
  table is organized as follows :
  - the table consists of chunks of 7 words (struct tos_symbol).
  - only type values of a200 and a248 have been found in studied samples. 
    they correspond to ordinary symbol whose sym_value is an offset from
    the start of the file once loaded (ie not counting the header).
  - if the symbol name is strictly less then 8 chars, then sym_name contains
    the name followed by 0, and sym_type is a200.
  - if the symbol name is exactly 8 char long, then sym_name contains
    the name, and sym_type is a200.
  - if the symbol name is strictly more then 8 chars, then sym_name contains
    the first 8 chars of the name, and sym_type is a248. The next chunk 
    contains the rest of the name in sym_name, with sym_type and sym_value 
    equal to zero.
  - I do not know if it is possible to represent a name longer than 16 chars,
    and if so how.

  */


